import json
import requests
from io import BytesIO
from zipfile import ZipFile
from pathlib import Path

__all__ = [
    'ZipUrl',
    'get_codemeta_from_zipurl',
]

class ZipUrl:

    def __init__(self, url):
        self.url = url
        self.content = requests.get(self.url)

    @property
    def filelist(self):
        with ZipFile(BytesIO(self.content.content)) as zipobj:
            return zipobj.namelist()

    def find_files(self, filename):
        """
        return the path of files in the archive matching `filename`

        :param filename: string
        :return: list[str]
        """
        matching_files = [f for f in self.filelist if Path(f).name == filename]
        if len(matching_files) == 0:
            raise FileNotFoundError(f"No file named {filename} in {self.url}")
        else:
            return matching_files

    def extract_file(self, filepath):
        with ZipFile(BytesIO(self.content.content)) as zipobj:
            return zipobj.extract(filepath)


def get_codemeta_from_zipurl(url):
    """
    Extract and reads codemeta metadata from a zip url.
    A codemeta.json file must be present in the zip archive.

    :param url: string
        url to a zip file
    :return: dictionnary
        metadata in the codemeta.json file in the zip archive
    """
    zipurl = ZipUrl(url)
    codemeta_paths = zipurl.find_files('codemeta.json')
    # if there are more than one codemeta file in the archive, we consider the one in the root directory, hence the
    # one with the shortest path
    codemeta_path = min(codemeta_paths, key=len)
    with open(zipurl.extract_file(codemeta_path)) as file:
        codemeta = json.load(file)

    return codemeta

